@@ -3,6 +3,7 @@ import warnings
 import time
 import logging
 import threading
+import operator
 
 from typing import Union, MutableMapping, List, Callable, Type, Optional, Iterable
 from functools import wraps
@@ -216,6 +217,29 @@ class ComponentState(InstanceRegistry):
     def current_operational_state(self, value: Union[int, str, StateSet]):
         self.update_operational_state(value)
 
+    def _get_state_from_key(self, value: Union[int, str, StateSet], state_set: str) -> StateSet:
+        """
+        Gets an operational state from the provided key. If an StateSet instance which is not the same
+        class as the assigned operational state class of the instance, the equivalent state will be retrieved.
+
+        :param value: value, name, or StateSet
+        :param state_set: state set key to use for retrieval
+        """
+        state_set = getattr(self, state_set)
+        if isinstance(value, state_set):
+            return value
+        elif isinstance(value, StateSet):
+            try:
+                return state_set(value.value)
+            except ValueError as e:
+                raise ValueError(f'the operational state {value} does not have an equivalent {state_set.__name__} value')
+        elif type(value) is str:
+            return state_set[value.upper()]
+        elif type(value) is int or isinstance(value, State):
+            return state_set(value)
+        else:
+            raise TypeError(f'the type {type(value)} is not supported for a current component operational state')
+
     def get_operational_state_from_key(self, value: Union[int, str, StateSet]) -> StateSet:
         """
         Gets an operational state from the provided key. If an StateSet instance which is not the same
@@ -223,19 +247,22 @@ class ComponentState(InstanceRegistry):
 
         :param value: value, name, or StateSet
         """
-        if isinstance(value, self.operational_states):
-            return value
-        elif isinstance(value, StateSet):
-            try:
-                return self.operational_states(value.value)
-            except ValueError as e:
-                raise ValueError(f'the operational state {value} does not have an equivalent {self.operational_states.__name__} value')
-        elif type(value) is str:
-            return self.operational_states[value.upper()]
-        elif type(value) is int or isinstance(value, State):
-            return self.operational_states(value)
-        else:
-            raise TypeError(f'the type {type(value)} is not supported for a current component operational state')
+        return self._get_state_from_key(
+            value,
+            'operational_states',
+        )
+
+    def get_action_state_from_key(self, value: Union[int, str, StateSet]) -> StateSet:
+        """
+        Gets an action state from the provided key. If an StateSet instance which is not the same
+        class as the assigned operational state class of the instance, the equivalent state will be retrieved.
+
+        :param value: value, name, or StateSet
+        """
+        return self._get_state_from_key(
+            value,
+            'action_states',
+        )
 
     def update_operational_state(self, state: Union[int, str, StateSet]) -> Optional[List]:
         """
@@ -431,6 +458,7 @@ class ComponentState(InstanceRegistry):
                                             target_flag_state=True,
                                             cycle_time: float = 0.1,
                                             threshold_state: Union[int, StateSet] = OperationalState.OFFLINE,
+                                            timeout: float = None,
                                             **flag_check_kwargs
                                             ):
         """
@@ -443,6 +471,8 @@ class ComponentState(InstanceRegistry):
         :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
         :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
         :param target_flag_state: target flag state to monitor for
+        :param timeout: optional timeout (seconds) to use while monitoring flags. If the timeout is reached before the
+            flag state is reached, an error is raised.
         :param flag_check_kwargs: keyword arguments for the flag check
         """
         # retrieve components as specified
@@ -451,11 +481,17 @@ class ComponentState(InstanceRegistry):
             f'waiting for {flag_check.__name__} to be {target_flag_state} while monitoring '
             + f'{len(monitored_components)} components' if len(monitored_components) > 0
             else monitored_components[0].component_name
+            + f' (timeout: {timeout} s)' if timeout is not None else ''
         )
+        if timeout is not None:
+            timeout = time.time() + timeout
         while flag_check(**flag_check_kwargs) != target_flag_state:
+            if timeout is not None and time.time() > timeout:
+                raise TimeoutError(
+                    f'the desired flag state "{target_flag_state}" was not encountered before the timeout elapsed '
+                )
             cls.check_component_states_raise(threshold_state, *monitored_components)
             time.sleep(cycle_time)
-            # todo add timeout
         logging.info(f'finished waiting, continuing')
 
     def state_monitor_flag_check(self,
@@ -463,6 +499,7 @@ class ComponentState(InstanceRegistry):
                                  target_flag_state=True,
                                  cycle_time: float = 0.1,
                                  threshold_state: Union[int, StateSet] = OperationalState.OFFLINE,
+                                 timeout: float = None,
                                  **flag_check_kwargs
                                  ):
         """
@@ -474,6 +511,8 @@ class ComponentState(InstanceRegistry):
         :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
         :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
         :param target_flag_state: target flag state to monitor for
+        :param timeout: optional timeout (seconds) to use while monitoring flags. If the timeout is reached before the
+            flag state is reached, an error is raised.
         :param flag_check_kwargs: keyword arguments for the flag check
         """
         self.components_state_monitor_flag_check(
@@ -482,6 +521,7 @@ class ComponentState(InstanceRegistry):
             target_flag_state=target_flag_state,
             cycle_time=cycle_time,
             threshold_state=threshold_state,
+            timeout=timeout,
             **flag_check_kwargs,
         )
 
@@ -522,22 +562,54 @@ class ComponentState(InstanceRegistry):
     def wait_for_component_operational_state(self,
                                              desired_state: Union[int, StateSet, str],
                                              cycle_time: float = 0.1,
-                                             timeout: float = None,  # todo implement timeout for wait
+                                             timeout: float = None,
+                                             comparison_operator: Callable = operator.ge,
                                              ):
         """
         Waits for the specified components to be at or above the specified operational state.
 
-        :param desired_state: desired state for the component to be in
+        :param desired_state: desired operational state for the component to be in
         :param cycle_time: state check cycle time
+        :param timeout: optional timeout while waiting for the operational state
+        :param comparison_operator: optional override for the comparison operator used by the wait method. The operator
+            should expect two arguments: the current and desired states respectively. The operator function should return
+            False if the condition is not met and True if it is.
         """
         # todo create a generic method to wait for multiple components
         desired_state = self.get_operational_state_from_key(desired_state)
         if timeout is not None:
             max_time = time.time() + timeout
         self.logger.info(f'waiting for {desired_state.name} operational state')
-        while self.current_operational_state < desired_state:
+        while comparison_operator(self.current_operational_state, desired_state) is False:
             if timeout is not None and time.time() > max_time:
-                self.logger.error(
-                    f'timeout of {timeout} s elapsed while waiting for {self.component_name} to be {desired_state.name}'
-                )
+                msg = f'timeout of {timeout} s elapsed while waiting for {self.component_name} to be {desired_state.name}'
+                self.logger.error(msg)
+                raise TimeoutError(msg)
+            time.sleep(cycle_time)
+
+    def wait_for_component_action_state(self,
+                                        desired_state: Union[int, StateSet, str],
+                                        cycle_time: float = 0.1,
+                                        timeout: float = None,
+                                        comparison_operator: Callable = operator.ge,
+                                        ):
+        """
+        Waits for the specified components to be at or above the specified action state.
+
+        :param desired_state: desired action state for the component to be in
+        :param cycle_time: state check cycle time
+        :param comparison_operator: optional override for the comparison operator used by the wait method. The operator
+            should expect two arguments: the current and desired states respectively. The operator function should return
+            False if the condition is not met and True if it is.
+        """
+        # todo create a generic method to wait for multiple components
+        desired_state = self.get_action_state_from_key(desired_state)
+        if timeout is not None:
+            max_time = time.time() + timeout
+        self.logger.info(f'waiting for {desired_state.name} action state')
+        while comparison_operator(self.current_action_state, desired_state) is False:
+            if timeout is not None and time.time() > max_time:
+                msg = f'timeout of {timeout} s elapsed while waiting for {self.component_name} to be {desired_state.name}'
+                self.logger.error(msg)
+                raise TimeoutError(msg)
             time.sleep(cycle_time)
